#ifdef GL_ES
precision highp float;
#endif


varying vec2 vtex0;
varying vec2 vtex1;
varying highp vec3 world_position;
varying highp vec3 world_normal;
varying highp vec3 world_bitangent;
varying highp vec3 world_tangent;

uniform highp vec3 light_dir[kMaxLights];
uniform highp vec4 light_diffuse[kMaxLights];
uniform highp vec4 light_ambient[kMaxLights];
uniform highp vec4 light_specular[kMaxLights];

uniform highp vec3 eye_pos_w;

const float PI = 3.14159265359;

uniform float metallic_param;
uniform float roughness_param;

uniform sampler2D diffuse_map;
uniform sampler2D normal_map;
uniform sampler2D metallic_map;
uniform sampler2D roughness_map;
#ifdef OCCLUSION
uniform float occlusion_param;
uniform sampler2D occlusion_map;
#endif
#ifdef REFLECTION
uniform sampler2D reflection_map;
#endif
#ifdef SUBSURFACE
uniform float subsurface_param;
uniform sampler2D subsurface_map;
#endif
#ifdef EMISSION
uniform float emission_param;
uniform sampler2D emission_map;
#endif
#ifdef LIGHTMAP
uniform float lmap_gamma;
uniform float lmap_mult;
uniform float lmap_add;
uniform sampler2D light_map;
uniform sampler2D shadow_map;
#endif
#ifdef FOG
varying float vfog;
uniform vec2  fogdist;
uniform vec3  fogcolor;
#endif


float SchlickFresnel (float u)
{
    float m = clamp (1.0 - u, 0.0, 1.0);
    float m2 = m * m;
    return m2 * m2 * m; // pow(m,5)
}



void main()
{
    highp vec4 diffuseColor = texture2D (diffuse_map, vtex0);
    highp vec3 diffuseColorLin = pow (diffuseColor.rgb, vec3 (2.2)); // srgb to linear

    float roughness = roughness_param * texture2D (roughness_map, vtex0).r;
    float r2 = roughness * roughness;
    float metallic = metallic_param * texture2D (metallic_map,  vtex0).r;
    float nonmetallic = 1.0 - metallic;
    highp vec3 nonmetallicColorLin = diffuseColorLin * nonmetallic;

    highp vec3 finalColor = vec3 (0.0);

    highp vec3 N = world_normal;
#ifdef NORMALMAP
    highp vec3 nmap = texture2D (normal_map, vtex0).rgb;
    highp vec3 n = 2.0 * nmap - 1.0;
    N = normalize (n.x * world_tangent + n.y * world_bitangent + n.z * world_normal);
#endif // NORMALMAP

    highp vec3 V = normalize (eye_pos_w - world_position); // inverse view direction
    highp float NdotV = clamp (abs (dot (N, V)), 0.001, 1.0);

    // used in diffuse calc -> Fd
    float FV = SchlickFresnel (NdotV);

    // used in specular reflectance calc -> F_
    highp vec3 specularColor = mix (vec3(0.04), diffuseColorLin, metallic); 
    float reflectance = max (max (specularColor.r, specularColor.g), specularColor.b);
    highp vec3 reflectance90 = vec3 (clamp (reflectance * 25.0, 0.0, 1.0));

    // used in geometric occlusion calc -> G_
    float atten_v = 2.0 * NdotV / (NdotV + sqrt (r2 + (1.0 - r2) * (NdotV * NdotV)));

    highp vec3 reflectionContrib = vec3(0.0);
#ifdef REFLECTION
    highp vec3 ray = reflect (V, N);
    ray.z += 1.0;
    float p = 2. * sqrt (ray.x * ray.x + ray.y * ray.y + ray.z * ray.z);
    // using mip-levels 0-4 for reflection focus, hence roughness * 4.0
    //highp vec3 reflectionColorLin = pow (texture2DLod (reflection_map, ray.xy / p + .5, roughness * 4.0).rgb, vec3 (2.2));
    highp vec3 reflectionColorLin = pow (texture2D (reflection_map, ray.xy / p + .5).rgb, vec3 (2.2));
    reflectionContrib = reflectionColorLin * diffuseColorLin * (1.0 - roughness); 
#endif

#ifdef LIGHTMAP
    highp vec3 lmap = pow (texture2D (light_map, vtex1).rgb, vec3 (lmap_gamma));
    lmap += lmap_add;
    lmap *= lmap_mult;
    highp vec3 shadowMask = texture2D (shadow_map, vtex1).rrr;
#endif

    // light loop
    for (int i = 0; i < kMaxLights; ++i)
    {
        highp vec3 L = -light_dir[i]; // inverse light direction
        highp vec3 H = normalize (L + V);

        highp float NdotL = clamp (dot (N, L), 0.001, 1.0);
        highp float NdotH = clamp (dot (N, H), 0.0, 1.0);
        highp float LdotH = clamp (dot (L, H), 0.0, 1.0);
        highp float LdotH2 = LdotH * LdotH;
        highp float VdotH = clamp (dot (V, H), 0.0, 1.0);

        // diffuse fresnel
        float FL = SchlickFresnel (NdotL);
        float Fd90 = .5 + 2. * LdotH2 * roughness;
        float Fd = mix (1.0, Fd90, FL) * mix (1.0, Fd90, FV);

#ifdef SUBSURFACE
        // Based on Hanrahan-Krueger brdf
        float Fss90 = LdotH2 * roughness;
        float Fss = mix (1., Fss90, FL) * mix (1., Fss90, FV);
        float ss = clamp (1.25 * (Fss * (1./(NdotL + NdotV) - .5) + .5), 0.0, 1.0);
        float subsurface = subsurface_param * texture2D (subsurface_map, vtex0).r;
        Fd = mix (Fd, ss, subsurface);
#endif

        // specular reflection
        highp vec3 F_ = specularColor + (reflectance90 - specularColor) * SchlickFresnel (VdotH);

        // geometric occlusion
        float atten_l = 2.0 * NdotL / (NdotL + sqrt (r2 + (1.0 - r2) * (NdotL * NdotL)));
        float G_ = atten_l * atten_v;

        // microfacet distribution
        float d_ = (r2 * NdotH - NdotH) * NdotH + 1.0;
        float D_ = r2 / (PI * d_ * d_);

        highp vec3 specularContrib = F_ * G_ * D_ / (4.0 * NdotL * NdotV);
        highp vec3 diffuseContrib  = Fd * nonmetallicColorLin;
        highp vec3 ambientContrib  = diffuseContrib;
#ifdef OCCLUSION
        ambientContrib = mix (diffuseContrib, diffuseContrib * texture2D (occlusion_map, vtex0).r, occlusion_param);
#endif

        highp vec3 outcolor = diffuseContrib * light_diffuse[i].rgb;
        outcolor += specularContrib * light_specular[i].rgb;
        outcolor *= NdotL;
#ifdef LIGHTMAP
        outcolor *= shadowMask.rgb;
#endif
        outcolor += mix (ambientContrib, reflectionContrib, metallic) * light_ambient[i].rgb;
#ifdef LIGHTMAP
        outcolor *= lmap.rgb;
#endif

        finalColor += outcolor;
    }

#ifdef EMISSION
    float emissive = emission_param * texture2D (emission_map, vtex0).r;
    finalColor = mix (finalColor, diffuseColorLin, emissive);
#endif

    finalColor = pow (finalColor, vec3 (.454545));
#ifdef FOG
    float fog = clamp ((vfog - fogdist.x) / (fogdist.y - fogdist.x), 0.0, 1.0);
    finalColor = mix (finalColor, fogcolor, fog);
#endif
    gl_FragColor = vec4 (finalColor, diffuseColor.a) * global_color;
}


