#ifdef GL_ES
precision highp float;
#endif


varying vec2 vtex0;
varying vec2 vtex1;
varying highp vec3 world_position;
varying highp vec3 world_normal;
varying highp vec3 world_bitangent;
varying highp vec3 world_tangent;

uniform lowp float pbr_active_lights;
uniform highp mat4 pbr_light_data[kMaxPBRLights];
uniform highp float pbr_exposure;
uniform highp vec4 pbr_ambient;

uniform highp vec3 eye_pos_w;

const float PI = 3.14159265359;

uniform sampler2D diffuse_map;
uniform sampler2D normal_map;
uniform float metallic_param;
uniform sampler2D metallic_map;
uniform float roughness_param;
uniform sampler2D roughness_map;
#ifdef OCCLUSION
uniform float occlusion_param;
uniform sampler2D occlusion_map;
#endif
#ifdef SUBSURFACE
uniform float subsurface_param;
uniform sampler2D subsurface_map;
#endif
#ifdef EMISSION
uniform float emission_param;
uniform sampler2D emission_map;
#endif
#ifdef REFLECTION
uniform sampler2D reflection_map;
#endif
#ifdef LIGHTMAP
uniform sampler2D light_map;
uniform sampler2D shadow_map;
uniform highp vec2 pbr_lm_exposure;
#endif

#ifdef ALPHATEST
uniform float alphatest;
#endif

#ifdef FOG
varying float vfog;
uniform vec2 fogdist;
uniform vec3 fogcolor;
uniform vec2 fogheight;
#endif

#ifdef DEMATERIALIZING
uniform float elapsed_time;
uniform float dematerialization_time;
#endif



float SchlickFresnel (float u)
{
    float m = 1.0 - u;
    float m2 = m * m;
    return m2 * m2 * m; // pow(m,5)
}



void main()
{
#ifdef DEMATERIALIZING
    highp float t1 = 2.0 * (dematerialization_time - elapsed_time);
    highp float t2 = 0.40*(0.5-mod(t1 * 5.0, 1.0)) + t1;
    highp vec3 coords = vec3(mod(world_position.x * 10.0, 1.0), mod(world_position.y * 10.0, 1.0), mod(world_position.z * 10.0, 1.0));
    if(t2 < mod((3.2345 * coords.x + 0.367)*( 3.9824 * coords.y  + 0.2345) * (3.678 * coords.z + 0.7892), 1.0))
            discard;
#endif
    highp vec4 diffuseColor = texture2D (diffuse_map, vtex0);
#ifdef ALPHATEST
    if (diffuseColor.a < alphatest)
    {
        discard;
    }
#endif
    highp vec3 diffuseColorLin = pow (diffuseColor.rgb, vec3 (2.2)); // srgb to linear

    float roughness = texture2D (roughness_map, vtex0).r * roughness_param;
    float r2 = roughness * roughness;
    float r2_compl = 1.0 - r2;
    float metallic = texture2D (metallic_map, vtex0).r * metallic_param;
    float nonmetallic = 1.0 - metallic;
    highp vec3 nonmetallicColorLin = diffuseColorLin * nonmetallic;
#ifdef SUBSURFACE
    float subsurface = texture2D (subsurface_map, vtex0).r * subsurface_param;
#endif // SUBSURFACE

    highp vec3 finalColor = vec3 (0.0);

    highp vec3 N = world_normal;
#ifdef NORMALMAP
    highp vec3 nmap = texture2D (normal_map, vtex0).rgb;
    highp vec3 n = 2.0 * nmap - 1.0;
    N = normalize (n.x * world_tangent + n.y * world_bitangent + n.z * world_normal);
#endif // NORMALMAP

    highp vec3 ambientContrib = pbr_ambient.rgb * pbr_ambient.a;
#ifdef LIGHTMAP
    highp vec3 shadow = texture2D (shadow_map, vtex1).rgb;
    highp vec3 lightmapContrib = pow (texture2D (light_map, vtex1).rgb, vec3 (2.2)) * pbr_lm_exposure.x + vec3(pbr_lm_exposure.y);
    finalColor = lightmapContrib * nonmetallicColorLin;
#endif

    highp vec3 V = normalize (eye_pos_w - world_position); // inverse view direction
    highp float NdotV = clamp (abs (dot (N, V)), 0.001, 1.0);

    // used in diffuse calc -> Fd
    float FV = SchlickFresnel (NdotV);

    // used in specular reflectance calc -> F_
    highp vec3 specularColor = mix (vec3(0.04), diffuseColorLin, metallic); 
    float reflectance = max (max (specularColor.r, specularColor.g), specularColor.b);
    highp vec3 reflectance90 = vec3 (clamp (reflectance * 25.0, 0.0, 1.0));

    // used in geometric occlusion calc -> G_
    float atten_v = NdotV + sqrt (r2 + r2_compl * (NdotV * NdotV));

    highp vec3 reflectionContrib = vec3(0.0);
#ifdef REFLECTION
    highp vec3 ray = reflect (V, N);
    ray.z += 1.0;
    float p = 2. * sqrt (ray.x * ray.x + ray.y * ray.y + ray.z * ray.z);
    // using mip-levels 0-4 for reflection focus, hence roughness * 4.0
    //highp vec3 reflectionColorLin = pow (texture2DLod (reflection_map, ray.xy / p + .5, roughness * 4.0).rgb, vec3 (2.2));
    highp vec3 reflectionColorLin = pow (texture2D (reflection_map, ray.xy / p + .5).rgb, vec3 (2.2));
    reflectionContrib = reflectionColorLin * diffuseColorLin * (1.0 - roughness); 
#endif

    // light loop
    for (int i = 0; i < int(pbr_active_lights); ++i)
    {
        highp mat4 light_data               = pbr_light_data[i];
        highp vec4 light_pos_intensity      = light_data[0];
        highp vec4 light_dir_type           = light_data[1];
        highp vec4 light_diff_inner_cone    = light_data[2];
        highp vec4 light_spec_outer_cone    = light_data[3];

        float non_directional = 1.0 + min (0.0, light_dir_type.w);
        highp vec3 LD = -light_dir_type.xyz + non_directional * ((light_pos_intensity.xyz - world_position) + light_dir_type.xyz);
        highp float d = length (LD);
        highp float I = light_pos_intensity.w / max(0.01, d * d);

        highp vec3 L = normalize (LD); // inverse light direction
        highp vec3 H = normalize (L + V);

        highp float NdotL = clamp (dot (N, L), 0.001, 1.0);
        highp float NdotH = max (dot (N, H), 0.0);
        highp float LdotH = max (dot (L, H), 0.0);
        highp float LdotH2 = LdotH * LdotH;
        highp float VdotH = max (dot (V, H), 0.0);

        highp float cone = clamp((dot (L, -normalize (light_dir_type.xyz)) - light_spec_outer_cone.w) / (light_diff_inner_cone.w - light_spec_outer_cone.w), 0.0, 1.0);
        I *= 1.0 + max(0.0, light_dir_type.w) * (cone - 1.0);

        // diffuse fresnel
        float FL = SchlickFresnel (NdotL);
        float Fd90 = .5 + 2. * LdotH2 * roughness;
        float Fd = mix (1.0, Fd90, FL) * mix (1.0, Fd90, FV);

#ifdef SUBSURFACE
        // Based on Hanrahan-Krueger brdf
        float Fss90 = LdotH2 * roughness;
        float Fss = mix (1., Fss90, FL) * mix (1., Fss90, FV);
        float ss = clamp (1.25 * (Fss * (1./(NdotL + NdotV) - .5) + .5), 0.0, 1.0);
        Fd = mix (Fd, ss, subsurface);
#endif

        // specular reflection
        highp vec3 F_ = specularColor + (reflectance90 - specularColor) * SchlickFresnel (VdotH);

        // geometric occlusion
        float atten_l = NdotL + sqrt (r2 + r2_compl * (NdotL * NdotL));
        float G_ = 1.0 / (atten_l * atten_v);

        // microfacet distribution
        float d_ = (r2 * NdotH - NdotH) * NdotH + 1.0;
        float D_ = r2 / (PI * d_ * d_);

        highp vec3 specularContrib = F_ * G_ * D_;
        highp vec3 diffuseContrib  = Fd * nonmetallicColorLin;

        highp vec3 outcolor = diffuseContrib * light_diff_inner_cone.rgb;
        outcolor += specularContrib * light_spec_outer_cone.rgb;
        outcolor *= NdotL;

#ifdef LIGHTMAP
        outcolor *= shadow;
#endif

        finalColor += I * outcolor;
    }
    
#ifdef OCCLUSION
    ambientContrib *= (texture2D (occlusion_map, vtex0).r * occlusion_param);
#endif
    finalColor += mix (ambientContrib * nonmetallicColorLin, ambientContrib * reflectionContrib, metallic);

#ifdef EMISSION
    float emissive = texture2D (emission_map, vtex0).r * emission_param;
    finalColor = mix (finalColor, diffuseColorLin, emissive);
#endif

    // exposure and tonemapping
    vec3 x = max(vec3(0.0), pbr_exposure * finalColor.xyz - vec3(0.004));
    finalColor.xyz = (x * (6.2 * x + vec3(0.5))) / (x * (6.2 * x + vec3(1.7)) + vec3(0.06));

#ifdef FOG
    float fog = clamp ((vfog - fogdist.x) / (fogdist.y - fogdist.x), 0.0, 1.0);
    fog *= 1.0 - clamp((world_position.y - fogheight.x) / (fogheight.y - fogheight.x), 0.0, 1.0);
    finalColor = mix (finalColor, fogcolor, fog);
#endif
    gl_FragColor = vec4 (finalColor, diffuseColor.a) * global_color;
}
