
#ifdef GL_ES
precision mediump float;
#endif

varying highp vec2 vtex0;
#ifdef QUADMATERIAL
varying highp vec4 vtex1;
#endif
varying highp vec3 wNormal;
varying highp vec3 wTangent;
varying highp vec3 wBitangent;
varying highp vec3 wPosition;
#ifdef PORTAL
varying highp vec3 portalPos;
varying highp vec3 portalEyeDir;
#endif

uniform sampler2D diffuse;
uniform sampler2D normalmap;
uniform sampler2D specmap;
#ifdef REFLECTION
uniform sampler2D reflectionMap;
#endif
#ifdef REFRACTION
uniform sampler2D refractionMap;
#endif

#ifdef VERTEX_LIGHTING
varying highp vec3 irradiance1;
varying highp vec3 irradiance2;
varying highp vec3 irradiance3;
uniform sampler2D icmap;
#endif

#ifdef SHADOW_MAP
varying highp vec4 shadowmapUV;
uniform sampler2D shadowmap;
#endif

#ifdef RIM_LIGHT
varying highp vec3 light1Dir;
#endif

#ifdef AO_MAP
uniform sampler2D omap;
#endif

#ifdef SSS
uniform sampler2D sssmap;
#endif

#ifdef ALPHA_THRESHOLD
uniform float alpha_threshold;
#endif

uniform highp vec4 light_diffuse[kMaxLights];
uniform highp vec4 light_ambient[kMaxLights];
uniform highp vec4 light_specular[kMaxLights];
uniform highp vec3 light_dir[kMaxLights];

uniform highp vec3 eye_pos_w;

uniform highp float elapsed_time;
uniform highp float dematerialization_time;

#ifdef MORPH
uniform highp vec4 morphWeight;
#endif

float specPow(float a, float b)
{
    return a/((1. - b)*a + b);
}

// saturate
float sat(float src)
{
    return clamp(src, 0., 1.);
}

vec3 sat(vec3 src)
{
    return clamp(src, 0., 1.);
}

float rgbToLuma(vec3 color)
{
    return dot(color, vec3(0.299, 0.587, 0.114));
}

#ifdef REFRACTION
vec3 myRefract(vec3 incidence, vec3 normal, float etaRatio, out float w)
{
    float cosI = -dot(incidence, normal);
    float cosT2 = 1. - etaRatio*etaRatio*(1. - cosI*cosI);

    w = sat(1. + cosT2*4.);

    vec3 r = incidence*etaRatio + normal*(etaRatio*cosI - sqrt(abs(cosT2)));

    return r;
}

vec2 hemisphereMapUV(vec3 v)
{
    v.y = -v.y;
    return v.xy*.5 + .5;
}
#endif

void main()
{
#ifdef DEMATERIALIZING
    highp float t1 = 2.0 * (dematerialization_time - elapsed_time);
    highp float t2 = 0.40*(0.5-mod(t1 * 5.0, 1.0)) + t1;
    highp vec3 coords = vec3(mod(wPosition.x * 10.0, 1.0), mod(wPosition.y * 10.0, 1.0), mod(wPosition.z * 10.0, 1.0));
    if(t2 < mod((3.2345 * coords.x + 0.367)*( 3.9824 * coords.y  + 0.2345) * (3.678 * coords.z + 0.7892), 1.0))
        discard;
#endif
#ifdef PORTAL
    float lambda = ( -portalPos.z)/portalEyeDir.z;  // I think we can get rid of this divide
    vec3 ipos = portalPos + lambda*portalEyeDir;
    if ((portalPos.z <= 0.0) && ( portalEyeDir.z > 0.0|| abs(ipos.x) > 1.0 || abs(ipos.y) > 1.0) )
    {
        discard;
    }
#endif

#ifdef QUADMATERIAL
    vec4 specData = texture2D(specmap, vtex1.xy);
#else
    vec4 specData = texture2D(specmap, vtex0);
#endif

#if defined(OPACITY) || defined(GHOST)
    vec4 specData2 = texture2D(specmap, vtex1.zy);
#endif
#if defined(REFLECTION) | defined(REFRACTION)
    vec4 specData3 = texture2D(specmap, vtex1.xw);
#endif

    vec3 nmapData = texture2D(normalmap, vtex0).rgb;
#ifdef MORPH
    vec3 nN = nmapData;
    nmapData += (texture2D(normalmap, vtex0+vec2(.5, 0.)).rgb - nN)*morphWeight.x;
    nmapData += (texture2D(normalmap, vtex0+vec2(0., .5)).rgb - nN)*morphWeight.y;
    nmapData += (texture2D(normalmap, vtex0+vec2(.5, .5)).rgb - nN)*morphWeight.z;
#endif
    vec3 tNormal = nmapData*2. - 1.;
    vec3 n = tNormal.x * wTangent + tNormal.y * wBitangent + tNormal.z * wNormal;
    n = normalize(n);
    
    vec3 u = normalize(wPosition - eye_pos_w);
    highp vec4 textureColor = texture2D(diffuse, vtex0);
    textureColor.rgb *= textureColor.rgb;

    vec3 skylightDir = light_dir[0].xyz;

#ifndef VERTEX_LIGHTING
    float dpSky = -dot(skylightDir, n);
#endif

#ifdef RIM_LIGHT
    vec3 sunlightDir = light1Dir;
#else
    vec3 sunlightDir = light_dir[1].xyz;
#endif
    vec3 sunlightColor = light_diffuse[1].rgb;

    float dpSun = -dot(sunlightDir, n);

#ifdef SSS
    vec3 sss = texture2D(sssmap, vtex0).rgb;
    sss *= sss;
    vec3 halfSSS = sss*.5;
    vec3 rSSS_1 = 1./(halfSSS + 1.);

#ifndef VERTEX_LIGHTING
    vec3 ground = light_ambient[0].rgb*sat(-dpSky*rSSS_1 * 0.8 + halfSSS + 0.2);
    vec3 sky = light_diffuse[0].rgb*sat(dpSky*rSSS_1 * 0.4 + halfSSS + 0.6);
#endif
    vec3 sun = sunlightColor*sat(dpSun*rSSS_1 + halfSSS);
#else   // !SSS
#ifndef VERTEX_LIGHTING
    vec3 ground = light_ambient[0].rgb*sat(-dpSky * 0.8 + 0.2);
    vec3 sky = light_diffuse[0].rgb*sat(dpSky * 0.4 + 0.6);
#endif
    vec3 sun = sunlightColor*sat(dpSun);
#endif
    
    vec3 r = u - 2.*n*dot(n, u);
    
    float smoothness = specData.r;
    float clothness = sat(0.05 - smoothness)*20.;   // treat really rough surface like cloth
    float smoothness3 = smoothness*smoothness*smoothness;
    float smoothPower = smoothness3*smoothness3*180.+ 0.0001;

#ifdef METALNESS
    float metal = specData.g;
    float notMetal = 1. - metal;
    vec3 diffuseColor = textureColor.rgb * notMetal;
#else
    float metal = 0.;
    float notMetal = 1.;
    vec3 diffuseColor = textureColor.rgb;
#endif

#ifdef HAIR
    metal = 1. - specData.b;
    notMetal = specData.b;
#endif

    float fresnel = (specPow(sat(1.05 + dot(n, u)), mix(.7*notMetal, 1., clothness)*5.)+ .111)*.9;

#ifdef HAIR
    float hairColorFresnel = fresnel;
    fresnel = fresnel * (1. - specData.b*.25) + specData.b*.25;
    fresnel *= .6;  // adjust hair reflection brightness
#endif

#ifndef VERTEX_LIGHTING
    float skyCos = dot(r, skylightDir);
    float skySpec = specPow(sat(-skyCos*.4+.6), smoothPower)*sat((dpSky * 0.4 + 0.6)*3.);
    float groundSpec = specPow(sat(skyCos*0.8+0.2), smoothPower)*sat((-dpSky * 0.8 + 0.2)*3.);
#endif
    float sunSpec = sat(-dot(r, sunlightDir));

#if defined(ANISO) | defined(HAIR)
    vec3 wNormal2 = normalize(wNormal);
    vec3 hairTangent = normalize(wBitangent - wNormal2*(tNormal.y*.5 - 0.05));
    vec3 t2 = normalize(hairTangent + wNormal2*0.1);

    vec3 hSun = normalize(u + sunlightDir);
    float sCosA = dot(hairTangent, hSun);
    float sunSpecAniso = 1. - sCosA*sCosA;
    sunSpecAniso *= sunSpecAniso;
    sunSpecAniso *= sunSpecAniso;
    sunSpecAniso *= sunSpecAniso;
    sunSpec = mix(sunSpec, sunSpecAniso, specData.b);
#endif

    sunSpec = specPow(sunSpec, smoothPower)*sat(dpSun*8.);

    float specIntensity = (0.5 + smoothness3*8.5);

#ifdef HAIR
    float sCosB = dot(t2, hSun);
    float sunSpecAniso2 = specPow(1. - sCosB*sCosB, smoothPower*2.);
    specIntensity *= 0.5;
#endif

#ifdef SHADOW_MAP
    highp vec2 screenPos = gl_FragCoord.xy;
    highp vec2 offset = mod(screenPos - 0.5, 2.);
    offset.x = abs(offset.x - offset.y); // x ^= y

    highp vec2 shadowMapScale = vec2(1./1024., 1./1024.);

    highp vec2 smUV = shadowmapUV.xy/shadowmapUV.w + offset*shadowMapScale;
    highp vec2 offsets = vec2(.5, -1.5) * shadowMapScale;

    vec4 shadowSamples = vec4(texture2D(shadowmap, smUV + offsets.yx).r,
                              texture2D(shadowmap, smUV + offsets.xx).r,
                              texture2D(shadowmap, smUV + offsets.yy).r,
                              texture2D(shadowmap, smUV + offsets.xy).r)*shadowmapUV.w;

    float shadowVal = (float(shadowSamples.x >= shadowmapUV.z) + float(shadowSamples.y >= shadowmapUV.z)
                       + float(shadowSamples.z >= shadowmapUV.z) + float(shadowSamples.w >= shadowmapUV.z))*.25;
    sun *= shadowVal;
    sunSpec *= shadowVal;
#ifdef HAIR
    sunSpecAniso2 *= shadowVal;
#endif
#endif

#ifdef AO_MAP
    vec3 ao = texture2D(omap, vtex0).rbg;
#endif

#if !defined(SHADOW_MAP) & defined(ANIMATION_AO)
    vec3 l = -sunlightDir;
    vec3 sunIC = texture2D(icmap, vec2(dot(l, wTangent), dot(l, wBitangent))*.5 + .5).rbg*2. - 1./3.;
    float sunAO = 1.;
#ifdef AO_MAP
    sunAO = min(sunAO, sat(dot(min(ao.rbg+.08, vcol.rgb), sunIC) + .15));
#endif

    float sunSelfShadow = sat(-dot(sunlightDir, wNormal) + .5);
    sun *= sunSelfShadow;
#ifdef SSS
    vec4 shadowVal = smoothstep(.8, .9, sunAO + vec4(sss*.05, 0));
    sun *= shadowVal.rgb;
    sunSelfShadow *= shadowVal.a;
#else
    float shadowVal = smoothstep(.8, .9, sunAO);
    sun *= shadowVal;
    sunSelfShadow *= shadowVal;
#endif
    sunSpec *= sunSelfShadow;
#ifdef HAIR
    sunSpecAniso2 *= sunSelfShadow;
#endif

#endif

#ifdef VERTEX_LIGHTING
    vec3 ic = texture2D(icmap, nmapData.xy).rbg*2. - 1./3.;
#ifdef AO_MAP
#ifdef MORPH
    vec3 o = ao;
    o += (texture2D(omap, vtex0 + vec2(.5, 0.)).rbg - ao)*morphWeight.x;
    o += (texture2D(omap, vtex0 + vec2(0., .5)).rbg - ao)*morphWeight.y;
    o += (texture2D(omap, vtex0 + vec2(.5, .5)).rbg - ao)*morphWeight.z;
    ic *= o;
#else   // !MORPH
    ic *= ao;
#endif
#endif
    vec3 final_color = max(vec3(0, 0, 0), irradiance1*ic.r + irradiance2*ic.g + irradiance3*ic.b);
#ifdef SSS
    final_color = mix(final_color, (irradiance1 + irradiance2 + irradiance3)*(1./3.), sss);
#endif

    final_color = (final_color + sun)*diffuseColor;
    
    vec3 rAbs = abs(r);
    float specMod = max(max(rAbs.x, rAbs.y), rAbs.z);

#if defined(ANISO) | defined(HAIR)
    vec3 lv = vec3(0.7071, 0., 0.7071)*rgbToLuma(irradiance1) +
    vec3(-.5*0.7071, 0.866*0.7071, 0.7071)*rgbToLuma(irradiance2) +
    vec3(-.5*0.7071, -0.866*0.7071, 0.7071)*rgbToLuma(irradiance3);
    lv = normalize(wTangent*lv.x + wBitangent*lv.y + wNormal*lv.z);
    vec3 h = normalize(u - lv);
    float cosA = dot(hairTangent, h);
    float anisoSpecMod = 1. - cosA*cosA;
    anisoSpecMod *= anisoSpecMod;
    anisoSpecMod *= anisoSpecMod;
    anisoSpecMod *= anisoSpecMod;
    specMod = mix(specMod, anisoSpecMod, specData.b);
#endif

#ifdef HAIR
    float cosB = dot(t2, h);
    float anisoSpecMod2 = specPow(1. - cosB*cosB, smoothPower*2.);
#endif

    specMod = specPow(specMod, smoothPower);

    vec3 specIC = texture2D(icmap, vec2(dot(r, wTangent), dot(r, wBitangent))*.5 + .5).rbg*2. - 1./3.;
#ifdef AO_MAP
    specIC *= ao;
#endif
    vec3 specLight = max(vec3(0., 0., 0.), irradiance1*specIC.x + irradiance2*specIC.y + irradiance3*specIC.z);
    vec3 specResult = specLight*specMod + sunlightColor*sunSpec;
#else   // !VERTEX_LIGHTING
    vec3 final_color = (ground + sky + sun)*diffuseColor;
    vec3 specResult = skySpec*light_diffuse[0].rgb +
                        groundSpec*light_ambient[0].rgb +
                       sunSpec*sunlightColor;
#ifdef AO_MAP
    final_color *= dot(ao, vec3(1./3., 1./3., 1./3.));
#endif
#endif

#ifdef AO_MAP
    float dao = dot(ao, ao)*(1./3.);
    specIntensity *= mix(dao*dao*dao*dao, 1., sat(metal + clothness));
#endif

    specResult *= textureColor.rgb*(metal*1.5 + clothness*.8) + notMetal*.3*(1.+ smoothness);
    final_color += specResult*specIntensity*fresnel;

#ifdef HAIR
    final_color += textureColor.rgb*(specLight*anisoSpecMod2 + sunlightColor*sunSpecAniso2)*specIntensity*hairColorFresnel*4.*specData.g;
#endif

#ifdef GHOST
    float ghost = specData2.b;
    textureColor.a *= 1. - ghost + fresnel * 0.25 * ghost;
#endif

#ifdef REFRACTION
    float opacity = specData3.g;
    float w = 1.;
    vec3 ref = myRefract(u, n, 1.1, w);
    vec3 ref_color = texture2D(refractionMap, hemisphereMapUV(ref).xy).rgb * w;
    ref_color *= ref_color;
    final_color += final_color * opacity + ref_color * (1. - opacity);
#endif

#ifdef EMISSIVE
    final_color += textureColor.rgb * specData.b;
#endif

#ifdef REFLECTION
    vec3 refl = texture2D(reflectionMap, -r.xy*.5 + .5).rgb;
    final_color += refl * refl * (sat((smoothness - .8)*5.) * fresnel);
#endif

    final_color = sqrt(final_color);

#ifdef OPACITY
    float opacity = specData2.g;
    texture_color.a *= opacity;
#endif

#if defined(PREMULTALPHA) & defined(OPACITY)
    final_color *= 1. - opacity;
#endif

#ifdef ALPHA_THRESHOLD
    if (textureColor.a < alpha_threshold) {
        discard;
    }
#endif

    gl_FragColor = vec4(final_color, textureColor.a)*global_color;
}
